from reprit.base import generate_repr

from ground.core.hints import QuaternaryPointFunction
from ground.hints import Coordinate
from .exact import point_point_point as exact_point_point_point
from .plain import point_point_point as plain_point_point_point
from .robust import point_point_point as robust_point_point_point

PointPointPointTest = QuaternaryPointFunction[Coordinate]


class Context:
    __slots__ = '_point_point_point_test',

    def __init__(self, point_point_point_test: PointPointPointTest) -> None:
        self._point_point_point_test = point_point_point_test

    __repr__ = generate_repr(__init__,
                             with_module_name=True)

    @property
    def point_point_point_test(self) -> PointPointPointTest:
        return self._point_point_point_test


exact_context = Context(point_point_point_test=exact_point_point_point.test)
plain_context = Context(point_point_point_test=plain_point_point_point.test)
robust_context = Context(point_point_point_test=robust_point_point_point.test)



from abc import abstractmethod as _abstractmethod
from numbers import Real as _Real
from typing import (Sequence as _Sequence,
                    TypeVar as _TypeVar)

try:
    from typing import Protocol as _Protocol
except ImportError:
    from typing_extensions import Protocol as _Protocol

Coordinate = _TypeVar('Coordinate',
                      bound=_Real)


class Point(_Protocol[Coordinate]):
    """
    **Point** is a minimal element of the plane
    defined by pair of real numbers (called *point's coordinates*).

    Points considered to be sorted lexicographically,
    with ``x`` coordinate being compared first.
    """

    @_abstractmethod
    def __new__(cls, x: Coordinate, y: Coordinate) -> 'Point':
        """Constructs point given its coordinates."""

    @_abstractmethod
    def __eq__(self, other: 'Point') -> bool:
        """Checks if the point is equal to the other."""

    @_abstractmethod
    def __hash__(self) -> int:
        """Returns hash value of the point."""

    @_abstractmethod
    def __lt__(self, other: 'Point') -> bool:
        """Checks if the point is less than the other."""

    @_abstractmethod
    def __le__(self, other: 'Point') -> bool:
        """Checks if the point is less than or equal to the other."""

    @property
    @_abstractmethod
    def x(self) -> Coordinate:
        """First coordinate of the point."""

    @property
    @_abstractmethod
    def y(self) -> Coordinate:
        """Second coordinate of the point."""


class Box(_Protocol[Coordinate]):
    """
    **Box** is a limited closed region
    defined by axis-aligned rectangular contour.
    """

    @_abstractmethod
    def __new__(cls,
                min_x: Coordinate,
                max_x: Coordinate,
                min_y: Coordinate,
                max_y: Coordinate) -> 'Box':
        """Constructs box given its coordinates limits."""

    @_abstractmethod
    def __eq__(self, other: 'Box') -> bool:
        """Checks if the box is equal to the other."""

    @property
    @_abstractmethod
    def max_x(self) -> Coordinate:
        """Returns maximum ``x``-coordinate of the box."""

    @property
    @_abstractmethod
    def max_y(self) -> Coordinate:
        """Returns maximum ``y``-coordinate of the box."""

    @property
    @_abstractmethod
    def min_x(self) -> Coordinate:
        """Returns minimum ``x``-coordinate of the box."""

    @property
    @_abstractmethod
    def min_y(self) -> Coordinate:
        """Returns minimum ``y``-coordinate of the box."""


class Multipoint(_Protocol[Coordinate]):
    """
    **Multipoint** is a non-empty set of unique points.
    """

    @_abstractmethod
    def __new__(cls, points: _Sequence[Point]) -> 'Multipoint':
        """Constructs multipoint given its points."""

    @_abstractmethod
    def __eq__(self, other: 'Multipoint') -> bool:
        """Checks if the multipoint is equal to the other."""

    @property
    @_abstractmethod
    def points(self) -> _Sequence[Point]:
        """Returns points of the multipoint."""


class Segment(_Protocol[Coordinate]):
    """
    **Segment** (or **line segment**)
    is a limited continuous part of the line containing more than one point
    defined by a pair of unequal points (called *segment's endpoints*).
    """

    @_abstractmethod
    def __new__(cls, start: Point, end: Point) -> 'Segment':
        """Constructs segment given its endpoints."""

    @_abstractmethod
    def __eq__(self, other: 'Segment') -> bool:
        """Checks if the segment is equal to the other."""

    @property
    @_abstractmethod
    def start(self) -> Point:
        """Returns start endpoint of the segment."""

    @property
    @_abstractmethod
    def end(self) -> Point:
        """Returns start endpoint of the segment."""


class Multisegment(_Protocol[Coordinate]):
    """
    **Multisegment**
    is a non-empty set of non-crossing and non-overlapping segments.
    """

    @_abstractmethod
    def __new__(cls, segments: _Sequence[Segment]) -> 'Multisegment':
        """Constructs multisegment given its segments."""

    @_abstractmethod
    def __eq__(self, other: 'Multisegment') -> bool:
        """Checks if the multisegment is equal to the other."""

    @property
    @_abstractmethod
    def segments(self) -> _Sequence[Segment]:
        """Returns segments of the multisegment."""


class Contour(_Protocol[Coordinate]):
    """
    **Contour** is a closed simple polyline defined by a sequence of points
    (called *contour's vertices*).
    """

    @_abstractmethod
    def __new__(cls, vertices: _Sequence[Point]) -> 'Contour':
        """Constructs contour given its vertices."""

    @_abstractmethod
    def __eq__(self, other: 'Contour') -> bool:
        """Checks if the contour is equal to the other."""

    @property
    @_abstractmethod
    def vertices(self) -> _Sequence[Point]:
        """Returns coordinates of the contour."""


class Polygon(_Protocol[Coordinate]):
    """
    **Polygon** is a limited closed region defined by the pair of outer contour
    (called *polygon's border*) and possibly empty sequence of inner contours
    (called *polygon's holes*).
    """

    @_abstractmethod
    def __new__(cls, border: Contour, holes: _Sequence[Contour]) -> 'Polygon':
        """Constructs polygon given its border and holes."""

    @_abstractmethod
    def __eq__(self, other: 'Polygon') -> bool:
        """Checks if the polygon is equal to the other."""

    @property
    @_abstractmethod
    def border(self) -> Contour:
        """Returns border of the polygon."""

    @property
    @_abstractmethod
    def holes(self) -> _Sequence[Contour]:
        """Returns holes of the polygon."""


class Multipolygon(_Protocol[Coordinate]):
    """
    **Multipolygon** is a non-empty set of non-overlapping polygons
    intersecting only in discrete set of points.
    """

    @_abstractmethod
    def __new__(cls, polygons: _Sequence[Polygon]) -> 'Multipolygon':
        """Constructs multipolygon given its polygons."""

    @_abstractmethod
    def __eq__(self, other: 'Multipolygon') -> bool:
        """Checks if the multipolygon is equal to the other."""

    @property
    @_abstractmethod
    def polygons(self) -> _Sequence[Polygon]:
        """Returns polygons of the multipolygon."""



from ground.hints import (Coordinate,
                          Point)
from .enums import (Kind,
                    Orientation)
from .hints import QuaternaryPointFunction


def kind(dot_producer: QuaternaryPointFunction,
         vertex: Point,
         first_ray_point: Point,
         second_ray_point: Point) -> Kind:
    return Kind(to_sign(dot_producer(vertex, first_ray_point, vertex,
                                     second_ray_point)))


def orientation(cross_producer: QuaternaryPointFunction,
                vertex: Point,
                first_ray_point: Point,
                second_ray_point: Point) -> Orientation:
    return Orientation(to_sign(cross_producer(vertex, first_ray_point, vertex,
                                              second_ray_point)))


def to_sign(value: Coordinate) -> int:
    return (1 if value > 0 else -1) if value else 0



from fractions import Fraction
from typing import Type

from ground.core.hints import QuaternaryPointFunction
from ground.core.linear.plain import segment as plain_segment
from ground.core.vector.plain.cross import multiply as plain_cross_product
from ground.hints import Point

contains_point = plain_segment.contains_point


def intersect(cross_product: QuaternaryPointFunction,
              point_cls: Type[Point],
              first_start: Point,
              first_end: Point,
              second_start: Point,
              second_end: Point) -> Point:
    if contains_point(cross_product, first_start, first_end, second_start):
        return second_start
    elif contains_point(cross_product, first_start, first_end, second_end):
        return second_end
    elif contains_point(cross_product, second_start, second_end, first_start):
        return first_start
    elif contains_point(cross_product, second_start, second_end, first_end):
        return first_end
    else:
        def to_exact_point(point: Point) -> Point:
            return point_cls(Fraction(point.x), Fraction(point.y))

        first_start, first_end, second_start, second_end = (
            to_exact_point(first_start), to_exact_point(first_end),
            to_exact_point(second_start), to_exact_point(second_end))
        scale = (plain_cross_product(first_start, second_start, second_start,
                                     second_end)
                 / plain_cross_product(first_start, first_end, second_start,
                                       second_end))
        return point_cls(first_start.x + (first_end.x - first_start.x) * scale,
                         first_start.y + (first_end.y - first_start.y) * scale)


relate = plain_segment.relate



from fractions import Fraction
from typing import (Callable,
                    Sequence,
                    Type)

from ground.hints import Point


def centroid(point_cls: Type[Point],
             points: Sequence[Point],
             inverse: Callable[[int], Fraction] = Fraction(1).__truediv__
             ) -> Point:
    result_x = result_y = 0
    for point in points:
        result_x += Fraction(point.x)
        result_y += Fraction(point.y)
    inverted_points_count = inverse(len(points))
    return point_cls(result_x * inverted_points_count,
                     result_y * inverted_points_count)



from ground.core.shewchuk import (sum_expansions,
                                  to_dot_product,
                                  two_mul,
                                  two_sub_tail,
                                  two_two_sum)
from ground.hints import (Coordinate,
                          Point)
from . import bounds


def multiply(first_start: Point,
             first_end: Point,
             second_start: Point,
             second_end: Point) -> Coordinate:
    first_start_x, first_start_y = first_start.x, first_start.y
    first_end_x, first_end_y = first_end.x, first_end.y
    second_start_x, second_start_y = second_start.x, second_start.y
    second_end_x, second_end_y = second_end.x, second_end.y
    addend_x = (first_end_x - first_start_x) * (second_end_x - second_start_x)
    addend_y = (first_end_y - first_start_y) * (second_end_y - second_start_y)
    result = addend_x + addend_y
    if 0 < addend_x:
        if 0 < addend_y:
            upper_bound = addend_x + addend_y
        else:
            return result
    elif addend_x < 0:
        if addend_y < 0:
            upper_bound = -addend_x - addend_y
        else:
            return result
    else:
        return result
    error_bound = bounds.to_multiply_first_error(upper_bound)
    if result >= error_bound or -result >= error_bound:
        return result
    return _adjusted_signed_area(first_start_x, first_start_y, first_end_x,
                                 first_end_y, second_start_x, second_start_y,
                                 second_end_x, second_end_y, upper_bound)


def _adjusted_signed_area(first_start_x: Coordinate,
                          first_start_y: Coordinate,
                          first_end_x: Coordinate,
                          first_end_y: Coordinate,
                          second_start_x: Coordinate,
                          second_start_y: Coordinate,
                          second_end_x: Coordinate,
                          second_end_y: Coordinate,
                          upper_bound: Coordinate) -> Coordinate:
    first_dx_head = first_end_x - first_start_x
    second_dx_head = second_end_x - second_start_x
    first_dy_head = first_end_y - first_start_y
    second_dy_head = second_end_y - second_start_y
    addend_x_tail, addend_x_head = two_mul(first_dx_head, second_dx_head)
    addend_y_tail, addend_y_head = two_mul(first_dy_head, second_dy_head)
    result_expansion = two_two_sum(addend_x_tail, addend_x_head,
                                   addend_y_tail, addend_y_head)
    result = sum(result_expansion)
    error_bound = bounds.to_multiply_second_error(upper_bound)
    if result >= error_bound or -result >= error_bound:
        return result
    first_dx_tail = two_sub_tail(first_end_x, first_start_x, first_dx_head)
    second_dx_tail = two_sub_tail(second_end_x, second_start_x, second_dx_head)
    first_dy_tail = two_sub_tail(first_end_y, first_start_y, first_dy_head)
    second_dy_tail = two_sub_tail(second_end_y, second_start_y, second_dy_head)
    if not (first_dx_tail or second_dx_tail or second_dy_tail
            or first_dy_tail):
        return result
    error_bound = (bounds.to_multiply_third_error(upper_bound)
                   + bounds.to_determinant_error(result))
    result += ((first_dx_head * second_dx_tail
                + second_dx_head * first_dx_tail)
               + (first_dy_head * second_dy_tail
                  + second_dy_head * first_dy_tail))
    if result >= error_bound or -result >= error_bound:
        return result
    result_expansion = sum_expansions(
            result_expansion, to_dot_product(first_dx_tail, first_dy_tail,
                                             second_dx_head, second_dy_head))
    result_expansion = sum_expansions(
            result_expansion, to_dot_product(first_dx_head, first_dy_head,
                                             second_dx_tail, second_dy_tail))
    result_expansion = sum_expansions(
            result_expansion, to_dot_product(first_dx_tail, first_dy_tail,
                                             second_dx_tail, second_dy_tail))
    return result_expansion[-1]



from typing import Type

from ground.hints import Box


def merge(box_cls: Type[Box], first_box: Box, second_box: Box) -> Box:
    return box_cls(min(first_box.min_x, second_box.min_x),
                   max(first_box.max_x, second_box.max_x),
                   min(first_box.min_y, second_box.min_y),
                   max(first_box.max_y, second_box.max_y))



from ground.hints import (Coordinate,
                          Point)


def test(first: Point,
         second: Point,
         third: Point,
         fourth: Point) -> Coordinate:
    first_dx, first_dy = first.x - fourth.x, first.y - fourth.y
    second_dx, second_dy = second.x - fourth.x, second.y - fourth.y
    third_dx, third_dy = third.x - fourth.x, third.y - fourth.y
    return ((first_dx * first_dx + first_dy * first_dy)
            * (second_dx * third_dy - second_dy * third_dx)
            - (second_dx * second_dx + second_dy * second_dy)
            * (first_dx * third_dy - first_dy * third_dx)
            + (third_dx * third_dx + third_dy * third_dy)
            * (first_dx * second_dy - first_dy * second_dx))



from fractions import Fraction
from typing import (Callable,
                    Sequence,
                    Type)

from ground.core.shewchuk import (scale_expansion,
                                  sum_expansions,
                                  to_cross_product)
from ground.hints import Point


def centroid(point_cls: Type[Point],
             vertices: Sequence[Point],
             inverse: Callable[[int], Fraction] = Fraction(1).__truediv__
             ) -> Point:
    double_area = x_numerator = y_numerator = (0,)
    prev = vertices[-1]
    prev_x, prev_y = prev.x, prev.y
    for vertex in vertices:
        x, y = vertex.x, vertex.y
        area_component = to_cross_product(prev_x, prev_y, x, y)
        double_area = sum_expansions(double_area, area_component)
        x_numerator = sum_expansions(x_numerator,
                                     scale_expansion(area_component,
                                                     prev_x + x))
        y_numerator = sum_expansions(y_numerator,
                                     scale_expansion(area_component,
                                                     prev_y + y))
        prev_x, prev_y = x, y
    inverted_denominator = inverse(3 * double_area[-1])
    return point_cls(x_numerator[-1] * inverted_denominator,
                     y_numerator[-1] * inverted_denominator)



from itertools import groupby
from typing import (Iterable,
                    List,
                    Sequence)

from ground.hints import Point
from .enums import Orientation
from .hints import (Domain,
                    TernaryPointFunction)


def to_convex_hull(orientation: TernaryPointFunction[Orientation],
                   points: Sequence[Point]) -> List[Point]:
    points = _to_unique_just_seen(sorted(points))
    lower, upper = (_to_sub_hull(orientation, points),
                    _to_sub_hull(orientation, reversed(points)))
    return lower[:-1] + upper[:-1] or points


def _to_unique_just_seen(iterable: Iterable[Domain]) -> Sequence[Domain]:
    return [key for key, _ in groupby(iterable)]


def _to_sub_hull(orientation: TernaryPointFunction[Orientation],
                 points: Iterable[Point]) -> List[Point]:
    result = []
    for point in points:
        while len(result) >= 2:
            if (orientation(result[-2], result[-1], point)
                    is not Orientation.COUNTERCLOCKWISE):
                del result[-1]
            else:
                break
        result.append(point)
    return result



from typing import (Callable,
                    Sequence,
                    Type)

from reprit.base import generate_repr

from ground.hints import Point
from .exact import (contour as exact_contour,
                    multipoint as exact_multipoint)
from .plain import (contour as plain_contour,
                    multipoint as plain_multipoint)
from .robust import (contour as robust_contour,
                     multipoint as robust_multipoint)

ContourCentroid = MultipointCentroid = Callable[[Type[Point], Sequence[Point]],
                                                Point]


class Context:
    __slots__ = '_contour_centroid', '_multipoint_centroid'

    def __init__(self,
                 contour_centroid: ContourCentroid,
                 multipoint_centroid: MultipointCentroid) -> None:
        self._contour_centroid = contour_centroid
        self._multipoint_centroid = multipoint_centroid

    __repr__ = generate_repr(__init__,
                             with_module_name=True)

    @property
    def contour_centroid(self) -> ContourCentroid:
        return self._contour_centroid

    @property
    def multipoint_centroid(self) -> MultipointCentroid:
        return self._multipoint_centroid


exact_context = Context(contour_centroid=exact_contour.centroid,
                        multipoint_centroid=exact_multipoint.centroid)
plain_context = Context(contour_centroid=plain_contour.centroid,
                        multipoint_centroid=plain_multipoint.centroid)
robust_context = Context(contour_centroid=robust_contour.centroid,
                         multipoint_centroid=robust_multipoint.centroid)



from typing import (Callable,
                    Sequence,
                    TypeVar)

from ground.hints import (Coordinate,
                          Point)

Expansion = Sequence[Coordinate]
Domain = TypeVar('Domain')
Range = TypeVar('Range')
QuaternaryPointFunction = Callable[[Point, Point, Point, Point], Range]
TernaryPointFunction = Callable[[Point, Point, Point], Range]



from typing import (Callable,
                    Type)

from reprit.base import generate_repr

from ground.core.enums import Relation
from ground.core.hints import QuaternaryPointFunction
from ground.hints import Point
from .exact import segment as exact_segment
from .plain import segment as plain_segment

ContainmentChecker = Callable[[QuaternaryPointFunction, Point, Point, Point],
                              bool]
Intersector = Callable[[QuaternaryPointFunction, Type[Point], Point, Point,
                        Point, Point], Point]
Relater = Callable[[QuaternaryPointFunction, Point, Point, Point, Point],
                   Relation]


class Context:
    __slots__ = '_containment_checker', '_intersector', '_relater'

    def __init__(self,
                 *,
                 intersector: Intersector,
                 containment_checker: ContainmentChecker,
                 relater: Relater) -> None:
        self._intersector, self._containment_checker, self._relater = (
            intersector, containment_checker, relater)

    __repr__ = generate_repr(__init__,
                             with_module_name=True)

    @property
    def intersector(self) -> Intersector:
        return self._intersector

    @property
    def containment_checker(self) -> ContainmentChecker:
        return self._containment_checker

    @property
    def relater(self) -> Relater:
        return self._relater


exact_context = Context(intersector=exact_segment.intersect,
                        containment_checker=exact_segment.contains_point,
                        relater=exact_segment.relate)
plain_context = Context(intersector=plain_segment.intersect,
                        containment_checker=plain_segment.contains_point,
                        relater=plain_segment.relate)



from numbers import Real

from ground.core.shewchuk import epsilon


def to_determinant_error(determinant: Real) -> Real:
    return (3 + 8 * epsilon) * epsilon * abs(determinant)


def to_multiply_first_error(upper_bound: Real) -> Real:
    return epsilon * (3 + 16 * epsilon) * upper_bound


def to_multiply_second_error(upper_bound: Real) -> Real:
    return epsilon * (2 + 12 * epsilon) * upper_bound


def to_multiply_third_error(upper_bound: Real) -> Real:
    return epsilon * epsilon * (9 + 64 * epsilon) * upper_bound



from typing import Tuple

from ground.hints import Coordinate
from .hints import Expansion


def _to_epsilon_and_splitter() -> Tuple[Coordinate, int]:
    every_other = True
    epsilon, splitter = 1., 1
    check = 1.
    while True:
        last_check = check
        epsilon /= 2.
        if every_other:
            splitter *= 2
        every_other = not every_other
        check = 1. + epsilon
        if check == 1. or check == last_check:
            break
    splitter += 1
    return epsilon, splitter


epsilon, splitter = _to_epsilon_and_splitter()


def fast_two_sum(left: Coordinate,
                 right: Coordinate) -> Tuple[Coordinate, Coordinate]:
    head = left + right
    right_virtual = head - left
    tail = right - right_virtual
    return tail, head


def two_sum(left: Coordinate,
            right: Coordinate) -> Tuple[Coordinate, Coordinate]:
    head = left + right
    right_virtual = head - left
    left_virtual = head - right_virtual
    right_tail = right - right_virtual
    left_tail = left - left_virtual
    tail = left_tail + right_tail
    return tail, head


def split(value: Coordinate,
          *,
          _splitter: Coordinate = splitter) -> Tuple[Coordinate, Coordinate]:
    base = _splitter * value
    high = base - (base - value)
    low = value - high
    return low, high


def two_mul(left: Coordinate,
            right: Coordinate) -> Tuple[Coordinate, Coordinate]:
    head = left * right
    left_low, left_high = split(left)
    right_low, right_high = split(right)
    first_error = head - left_high * right_high
    second_error = first_error - left_low * right_high
    third_error = second_error - left_high * right_low
    tail = left_low * right_low - third_error
    return tail, head


def two_mul_presplit(left: Coordinate,
                     right: Coordinate,
                     right_low: Coordinate,
                     right_high: Coordinate) -> Tuple[Coordinate, Coordinate]:
    head = left * right
    left_low, left_high = split(left)
    first_error = head - left_high * right_high
    second_error = first_error - left_low * right_high
    third_error = second_error - left_high * right_low
    tail = left_low * right_low - third_error
    return tail, head


def two_two_sub(left_tail: Coordinate,
                left_head: Coordinate,
                right_tail: Coordinate,
                right_head: Coordinate
                ) -> Tuple[Coordinate, Coordinate, Coordinate, Coordinate]:
    third_tail, mid_tail, mid_head = two_one_sub(left_tail, left_head,
                                                 right_tail)
    second_tail, first_tail, head = two_one_sub(mid_tail, mid_head, right_head)
    return third_tail, second_tail, first_tail, head


def two_two_sum(left_tail: Coordinate,
                left_head: Coordinate,
                right_tail: Coordinate,
                right_head: Coordinate
                ) -> Tuple[Coordinate, Coordinate, Coordinate, Coordinate]:
    third_tail, mid_tail, mid_head = two_one_sum(left_tail, left_head,
                                                 right_tail)
    second_tail, first_tail, head = two_one_sum(mid_tail, mid_head, right_head)
    return third_tail, second_tail, first_tail, head


def two_one_sum(left_tail: Coordinate,
                left_head: Coordinate,
                right: Coordinate
                ) -> Tuple[Coordinate, Coordinate, Coordinate]:
    second_tail, mid_head = two_sum(left_tail, right)
    first_tail, head = two_sum(left_head, mid_head)
    return second_tail, first_tail, head


def two_one_sub(left_tail: Coordinate,
                left_head: Coordinate,
                right: Coordinate
                ) -> Tuple[Coordinate, Coordinate, Coordinate]:
    second_tail, mid_head = two_sub(left_tail, right)
    first_tail, head = two_sum(left_head, mid_head)
    return second_tail, first_tail, head


def two_one_mul(left_tail: Coordinate,
                left_head: Coordinate,
                right: Coordinate
                ) -> Tuple[Coordinate, Coordinate, Coordinate, Coordinate]:
    right_low, right_high = split(right)
    head_tail, head = two_mul_presplit(left_head, right, right_low, right_high)
    tail_tail, tail_head = two_mul_presplit(left_tail, right, right_low,
                                            right_high)
    mid_tail, first_tail = two_sum(head_tail, tail_head)
    third_tail, second_tail = fast_two_sum(tail_tail, mid_tail)
    return third_tail, second_tail, first_tail, head


def two_sub(left: Coordinate,
            right: Coordinate) -> Tuple[Coordinate, Coordinate]:
    head = left - right
    return two_sub_tail(left, right, head), head


def two_sub_tail(left: Coordinate,
                 right: Coordinate,
                 head: Coordinate) -> Coordinate:
    right_virtual = left - head
    left_virtual = head + right_virtual
    right_error = right_virtual - right
    left_error = left - left_virtual
    return left_error + right_error


def square(value: Coordinate) -> Tuple[Coordinate, Coordinate]:
    head = value * value
    value_low, value_high = split(value)
    first_error = head - value_high * value_high
    second_error = first_error - (value_high + value_high) * value_low
    tail = value_low * value_low - second_error
    return tail, head


def add_to_expansion(expansion: Expansion, value: Coordinate) -> Expansion:
    """
    Adds given value to the expansion with zero components elimination.
    """
    result = []
    accumulator = value
    for index, component in enumerate(expansion):
        tail, accumulator = two_sum(accumulator, component)
        if tail:
            result.append(tail)
    if accumulator or not result:
        result.append(accumulator)
    return result


def sum_expansions(left: Expansion, right: Expansion) -> Expansion:
    """
    Sums two expansions with zero components elimination.
    """
    left_length, right_length = len(left), len(right)
    left_component, right_component = left[0], right[0]
    left_index = right_index = 0
    if ((right_component > left_component)
            is (right_component > -left_component)):
        accumulator = left_component
        left_index += 1
    else:
        accumulator = right_component
        right_index += 1
    result = []
    if (left_index < left_length) and (right_index < right_length):
        left_component, right_component = left[left_index], right[right_index]
        if ((right_component > left_component)
                is (right_component > -left_component)):
            tail, accumulator = fast_two_sum(left_component, accumulator)
            left_index += 1
        else:
            tail, accumulator = fast_two_sum(right_component, accumulator)
            right_index += 1
        if tail:
            result.append(tail)
        while (left_index < left_length) and (right_index < right_length):
            left_component, right_component = (left[left_index],
                                               right[right_index])
            if ((right_component > left_component)
                    is (right_component > -left_component)):
                tail, accumulator = two_sum(accumulator, left_component)
                left_index += 1
            else:
                tail, accumulator = two_sum(accumulator, right_component)
                right_index += 1
            if tail:
                result.append(tail)
    for left_index in range(left_index, left_length):
        left_component = left[left_index]
        tail, accumulator = two_sum(accumulator, left_component)
        if tail:
            result.append(tail)
    for right_index in range(right_index, right_length):
        right_component = right[right_index]
        tail, accumulator = two_sum(accumulator, right_component)
        if tail:
            result.append(tail)
    if accumulator or not result:
        result.append(accumulator)
    return result


def scale_expansion(expansion: Expansion, scalar: Coordinate) -> Expansion:
    """
    Multiplies the expansion by given scalar with zero components elimination.
    """
    components = iter(expansion)
    scalar_low, scalar_high = split(scalar)
    tail, accumulator = two_mul_presplit(next(components), scalar, scalar_low,
                                         scalar_high)
    result = []
    if tail:
        result.append(tail)
    for component in components:
        product_tail, product = two_mul_presplit(component, scalar, scalar_low,
                                                 scalar_high)
        tail, interim = two_sum(accumulator, product_tail)
        if tail:
            result.append(tail)
        tail, accumulator = fast_two_sum(product, interim)
        if tail:
            result.append(tail)
    if accumulator or not result:
        result.append(accumulator)
    return result


def to_cross_product(first_x: Coordinate,
                     first_y: Coordinate,
                     second_x: Coordinate,
                     second_y: Coordinate) -> Expansion:
    """
    Returns expansion of vectors' cross product.
    """
    minuend_tail, minuend_head = two_mul(first_x, second_y)
    subtrahend_tail, subtrahend_head = two_mul(second_x, first_y)
    return two_two_sub(minuend_tail, minuend_head, subtrahend_tail,
                       subtrahend_head)


def to_dot_product(first_x: Coordinate, first_y: Coordinate,
                   second_x: Coordinate, second_y: Coordinate) -> Expansion:
    """
    Returns expansion of vectors' dot product.
    """
    x_tail, x_head = two_mul(first_x, second_x)
    y_tail, y_head = two_mul(first_y, second_y)
    return two_two_sum(x_tail, x_head, y_tail, y_head)



from functools import reduce
from typing import (Iterable,
                    Tuple)

from ground.core.hints import Expansion
from ground.core.shewchuk import (scale_expansion,
                                  square,
                                  sum_expansions,
                                  to_cross_product,
                                  two_mul,
                                  two_sub_tail,
                                  two_two_sub,
                                  two_two_sum)
from ground.hints import (Coordinate,
                          Point)
from . import bounds


def test(first: Point,
         second: Point,
         third: Point,
         fourth: Point) -> Coordinate:
    first_x, first_y = first.x, first.y
    second_x, second_y = second.x, second.y
    third_x, third_y = third.x, third.y
    fourth_x, fourth_y = fourth.x, fourth.y
    first_dx, first_dy = first_x - fourth_x, first_y - fourth_y
    second_dx, second_dy = second_x - fourth_x, second_y - fourth_y
    third_dx, third_dy = third_x - fourth_x, third_y - fourth_y
    first_squared_distance = first_dx * first_dx + first_dy * first_dy
    second_squared_distance = second_dx * second_dx + second_dy * second_dy
    third_squared_distance = third_dx * third_dx + third_dy * third_dy
    first_dx_second_dy = first_dx * second_dy
    first_dx_third_dy = first_dx * third_dy
    second_dx_first_dy = second_dx * first_dy
    second_dx_third_dy = second_dx * third_dy
    third_dx_first_dy = third_dx * first_dy
    third_dx_second_dy = third_dx * second_dy
    result = (first_squared_distance
              * (second_dx_third_dy - third_dx_second_dy)
              + second_squared_distance
              * (third_dx_first_dy - first_dx_third_dy)
              + third_squared_distance
              * (first_dx_second_dy - second_dx_first_dy))
    upper_bound = (first_squared_distance
                   * (abs(second_dx_third_dy) + abs(third_dx_second_dy))
                   + second_squared_distance
                   * (abs(third_dx_first_dy) + abs(first_dx_third_dy))
                   + third_squared_distance
                   * (abs(first_dx_second_dy) + abs(second_dx_first_dy)))
    error_bound = bounds.to_point_point_point_first_error(upper_bound)
    if result > error_bound or -result > error_bound:
        return result
    return _adjusted_determinant(first_x, first_y, second_x, second_y, third_x,
                                 third_y, fourth_x, fourth_y, upper_bound)


def _adjusted_determinant(first_x: Coordinate,
                          first_y: Coordinate,
                          second_x: Coordinate,
                          second_y: Coordinate,
                          third_x: Coordinate,
                          third_y: Coordinate,
                          fourth_x: Coordinate,
                          fourth_y: Coordinate,
                          upper_bound: Coordinate) -> Coordinate:
    first_dx_head, first_dy_head = first_x - fourth_x, first_y - fourth_y
    second_dx_head, second_dy_head = second_x - fourth_x, second_y - fourth_y
    third_dx_head, third_dy_head = third_x - fourth_x, third_y - fourth_y
    first_second_cross_product = to_cross_product(
            first_dx_head, first_dy_head, second_dx_head, second_dy_head)
    second_third_cross_product = to_cross_product(
            second_dx_head, second_dy_head, third_dx_head, third_dy_head)
    third_first_cross_product = to_cross_product(third_dx_head, third_dy_head,
                                                 first_dx_head, first_dy_head)
    result_expansion = sum_expansions(
            sum_expansions(
                    _multiply_by_squared_length(second_third_cross_product,
                                                first_dx_head, first_dy_head),
                    _multiply_by_squared_length(third_first_cross_product,
                                                second_dx_head,
                                                second_dy_head)),
            _multiply_by_squared_length(first_second_cross_product,
                                        third_dx_head, third_dy_head))
    result = sum(result_expansion)
    error_bound = bounds.to_point_point_point_second_error(upper_bound)
    if result >= error_bound or -result >= error_bound:
        return result
    first_dx_tail = two_sub_tail(first_x, fourth_x, first_dx_head)
    first_dy_tail = two_sub_tail(first_y, fourth_y, first_dy_head)
    second_dx_tail = two_sub_tail(second_x, fourth_x, second_dx_head)
    second_dy_tail = two_sub_tail(second_y, fourth_y, second_dy_head)
    third_dx_tail = two_sub_tail(third_x, fourth_x, third_dx_head)
    third_dy_tail = two_sub_tail(third_y, fourth_y, third_dy_head)
    if not (first_dx_tail or first_dy_tail or second_dx_tail or second_dy_tail
            or third_dx_tail or third_dy_tail):
        return result
    error_bound = (bounds.to_point_point_point_third_error(upper_bound)
                   + bounds.to_point_point_point_determinant_error(result))
    result += (_to_addend(first_dx_head, first_dx_tail, first_dy_head,
                          first_dy_tail, second_dx_head, second_dx_tail,
                          second_dy_head, second_dy_tail, third_dx_head,
                          third_dx_tail, third_dy_head, third_dy_tail)
               + _to_addend(second_dx_head, second_dx_tail, second_dy_head,
                            second_dy_tail, third_dx_head, third_dx_tail,
                            third_dy_head, third_dy_tail, first_dx_head,
                            first_dx_tail, first_dy_head, first_dy_tail)
               + _to_addend(third_dx_head, third_dx_tail, third_dy_head,
                            third_dy_tail, first_dx_head, first_dx_tail,
                            first_dy_head, first_dy_tail, second_dx_head,
                            second_dx_tail, second_dy_head, second_dy_tail))
    if result >= error_bound or -result >= error_bound:
        return result
    first_squared_length = (_to_squared_length(first_dx_head, first_dy_head)
                            if (second_dx_tail or second_dy_tail
                                or third_dx_tail or third_dy_tail)
                            else (0,) * 4)
    second_squared_length = (_to_squared_length(second_dx_head, second_dy_head)
                             if (first_dx_tail or first_dy_tail
                                 or third_dx_tail or third_dy_tail)
                             else (0,) * 4)
    third_squared_length = (_to_squared_length(third_dx_head, third_dy_head)
                            if (first_dx_tail or first_dy_tail
                                or second_dx_tail or second_dy_tail)
                            else (0,) * 4)
    if first_dx_tail or first_dy_tail:
        if second_dx_tail or second_dy_tail or third_dx_tail or third_dy_tail:
            second_third_crossed_tails_tail, second_third_crossed_tails = (
                _to_crossed_tails(second_dx_head, second_dx_tail,
                                  second_dy_head, second_dy_tail,
                                  third_dx_head, third_dx_tail, third_dy_head,
                                  third_dy_tail))
        else:
            second_third_crossed_tails_tail = second_third_crossed_tails = (0,)
        if first_dx_tail:
            first_dx_tail_second_third_cross_product = scale_expansion(
                    second_third_cross_product, first_dx_tail)
            result_expansion = reduce(
                    sum_expansions,
                    _to_dx_extras(first_dx_tail_second_third_cross_product,
                                  first_dx_head, first_dx_tail, second_dy_tail,
                                  second_squared_length, third_dy_tail,
                                  third_squared_length,
                                  second_third_crossed_tails,
                                  second_third_crossed_tails_tail),
                    sum_expansions(
                            result_expansion,
                            _to_extra(first_dx_tail_second_third_cross_product,
                                      first_dx_head, first_dx_tail,
                                      second_dy_head, second_squared_length,
                                      third_dy_head, third_squared_length)))
        if first_dy_tail:
            first_dy_tail_second_third_cross_product = scale_expansion(
                    second_third_cross_product, first_dy_tail)
            result_expansion = reduce(
                    sum_expansions,
                    _to_dy_extras(first_dy_tail_second_third_cross_product,
                                  first_dy_head, first_dy_tail,
                                  second_third_crossed_tails,
                                  second_third_crossed_tails_tail),
                    sum_expansions(
                            result_expansion,
                            _to_extra(first_dy_tail_second_third_cross_product,
                                      first_dy_head, first_dy_tail,
                                      third_dx_head, third_squared_length,
                                      second_dx_head, second_squared_length)))
    if second_dx_tail or second_dy_tail:
        if first_dx_tail or first_dy_tail or third_dx_tail or third_dy_tail:
            third_first_crossed_tails_tail, third_first_crossed_tails = (
                _to_crossed_tails(third_dx_head, third_dx_tail, third_dy_head,
                                  third_dy_tail, first_dx_head, first_dx_tail,
                                  first_dy_head, first_dy_tail))
        else:
            third_first_crossed_tails_tail = third_first_crossed_tails = (0,)
        if second_dx_tail:
            second_dx_tail_third_first_cross_product = scale_expansion(
                    third_first_cross_product, second_dx_tail)
            result_expansion = reduce(
                    sum_expansions,
                    _to_dx_extras(second_dx_tail_third_first_cross_product,
                                  second_dx_head, second_dx_tail,
                                  third_dy_tail, third_squared_length,
                                  first_dy_tail, first_squared_length,
                                  third_first_crossed_tails,
                                  third_first_crossed_tails_tail),
                    sum_expansions(
                            result_expansion,
                            _to_extra(second_dx_tail_third_first_cross_product,
                                      second_dx_head, second_dx_tail,
                                      third_dy_head, third_squared_length,
                                      first_dy_head, first_squared_length)))
        if second_dy_tail:
            second_dy_tail_third_first_cross_product = scale_expansion(
                    third_first_cross_product, second_dy_tail)
            result_expansion = reduce(
                    sum_expansions,
                    _to_dy_extras(second_dy_tail_third_first_cross_product,
                                  second_dy_head, second_dy_tail,
                                  third_first_crossed_tails,
                                  third_first_crossed_tails_tail),
                    sum_expansions(
                            result_expansion,
                            _to_extra(second_dy_tail_third_first_cross_product,
                                      second_dy_head, second_dy_tail,
                                      first_dx_head, first_squared_length,
                                      third_dx_head, third_squared_length)))
    if third_dx_tail or third_dy_tail:
        if first_dx_tail or first_dy_tail or second_dx_tail or second_dy_tail:
            first_second_crossed_tails_tail, first_second_crossed_tails = (
                _to_crossed_tails(first_dx_head, first_dx_tail, first_dy_head,
                                  first_dy_tail, second_dx_head,
                                  second_dx_tail, second_dy_head,
                                  second_dy_tail))
        else:
            first_second_crossed_tails_tail = first_second_crossed_tails = (0,)
        if third_dx_tail:
            third_dx_tail_first_second_cross_product = scale_expansion(
                    first_second_cross_product, third_dx_tail)
            result_expansion = reduce(
                    sum_expansions,
                    _to_dx_extras(third_dx_tail_first_second_cross_product,
                                  third_dx_head, third_dx_tail, first_dy_tail,
                                  first_squared_length, second_dy_tail,
                                  second_squared_length,
                                  first_second_crossed_tails,
                                  first_second_crossed_tails_tail),
                    sum_expansions(
                            result_expansion,
                            _to_extra(third_dx_tail_first_second_cross_product,
                                      third_dx_head, third_dx_tail,
                                      first_dy_head, first_squared_length,
                                      second_dy_head, second_squared_length)))
        if third_dy_tail:
            third_dy_tail_first_second_cross_product = scale_expansion(
                    first_second_cross_product, third_dy_tail)
            result_expansion = reduce(
                    sum_expansions,
                    _to_dy_extras(third_dy_tail_first_second_cross_product,
                                  third_dy_head, third_dy_tail,
                                  first_second_crossed_tails,
                                  first_second_crossed_tails_tail),
                    sum_expansions(
                            result_expansion,
                            _to_extra(third_dy_tail_first_second_cross_product,
                                      third_dy_head, third_dy_tail,
                                      second_dx_head, second_squared_length,
                                      first_dx_head, first_squared_length)))
    return result_expansion[-1]


def _to_dx_extras(head: Expansion,
                  dx_head: Coordinate,
                  dx_tail: Coordinate,
                  left_dy_tail: Coordinate,
                  left_squared_length: Expansion,
                  right_dy_tail: Coordinate,
                  right_squared_length: Expansion,
                  left_right_crossed_tails_head: Expansion,
                  left_right_crossed_tails_tail: Expansion
                  ) -> Iterable[Expansion]:
    dx_tail_left_right_crossed_tails = scale_expansion(
            left_right_crossed_tails_head, dx_tail)
    yield sum_expansions(scale_expansion(head, dx_tail),
                         scale_expansion(dx_tail_left_right_crossed_tails,
                                         2 * dx_head))
    if left_dy_tail:
        yield scale_expansion(scale_expansion(right_squared_length, dx_tail),
                              left_dy_tail)
    if right_dy_tail:
        yield scale_expansion(scale_expansion(left_squared_length, -dx_tail),
                              right_dy_tail)
    first_addend = scale_expansion(dx_tail_left_right_crossed_tails, dx_tail)
    dx_tail_left_right_crossed_tails_tail = scale_expansion(
            left_right_crossed_tails_tail, dx_tail)
    second_addend = sum_expansions(
            scale_expansion(dx_tail_left_right_crossed_tails_tail,
                            2 * dx_head),
            scale_expansion(dx_tail_left_right_crossed_tails_tail, dx_tail))
    yield sum_expansions(first_addend, second_addend)


def _to_dy_extras(expansion: Expansion,
                  dy_head: Coordinate,
                  dy_tail: Coordinate,
                  rest_crossed_tails_head: Expansion,
                  rest_crossed_tails_tail: Expansion) -> Iterable[Expansion]:
    dy_tail_rest_crossed_tails = scale_expansion(rest_crossed_tails_head,
                                                 dy_tail)
    yield sum_expansions(scale_expansion(expansion, dy_tail),
                         scale_expansion(dy_tail_rest_crossed_tails,
                                         2 * dy_head))
    first_addend = scale_expansion(dy_tail_rest_crossed_tails, dy_tail)
    dy_tail_rest_crossed_tails_tail = scale_expansion(rest_crossed_tails_tail,
                                                      dy_tail)
    second_addend = sum_expansions(
            scale_expansion(dy_tail_rest_crossed_tails_tail, 2 * dy_head),
            scale_expansion(dy_tail_rest_crossed_tails_tail, dy_tail))
    yield sum_expansions(first_addend, second_addend)


def _to_extra(head: Expansion,
              coordinate: Coordinate,
              coordinate_tail: Coordinate,
              left_coordinate: Coordinate,
              left_squared_length: Expansion,
              right_coordinate: Coordinate,
              right_squared_length: Expansion) -> Expansion:
    second_addend = scale_expansion(scale_expansion(right_squared_length,
                                                    coordinate_tail),
                                    left_coordinate)
    first_addend = scale_expansion(head, 2 * coordinate)
    minuend = sum_expansions(first_addend, second_addend)
    subtrahend = scale_expansion(scale_expansion(left_squared_length,
                                                 coordinate_tail),
                                 -right_coordinate)
    return sum_expansions(subtrahend, minuend)


def _to_crossed_tails(left_dx_head: Coordinate,
                      left_dx_tail: Coordinate,
                      left_dy_head: Coordinate,
                      left_dy_tail: Coordinate,
                      right_dx_head: Coordinate,
                      right_dx_tail: Coordinate,
                      right_dy_head: Coordinate,
                      right_dy_tail: Coordinate
                      ) -> Tuple[Expansion, Expansion]:
    tail = two_two_sub(*two_mul(left_dx_tail, right_dy_tail),
                       *two_mul(right_dx_tail, left_dy_tail))
    head = sum_expansions(two_two_sum(*two_mul(left_dx_tail, right_dy_head),
                                      *two_mul(left_dx_head, right_dy_tail)),
                          two_two_sum(*two_mul(right_dx_tail, -left_dy_head),
                                      *two_mul(right_dx_head, -left_dy_tail)))
    return tail, head


def _multiply_by_squared_length(expansion: Expansion,
                                dx_head: Coordinate,
                                dy_head: Coordinate) -> Expansion:
    return sum_expansions(
            scale_expansion(scale_expansion(expansion, dx_head), dx_head),
            scale_expansion(scale_expansion(expansion, dy_head), dy_head))


def _to_addend(left_dx_head: Coordinate,
               left_dx_tail: Coordinate,
               left_dy_head: Coordinate,
               left_dy_tail: Coordinate,
               mid_dx_head: Coordinate,
               mid_dx_tail: Coordinate,
               mid_dy_head: Coordinate,
               mid_dy_tail: Coordinate,
               right_dx_head: Coordinate,
               right_dx_tail: Coordinate,
               right_dy_head: Coordinate,
               right_dy_tail: Coordinate) -> Coordinate:
    return ((left_dx_head * left_dx_head + left_dy_head * left_dy_head)
            * ((mid_dx_head * right_dy_tail + right_dy_head * mid_dx_tail)
               - (mid_dy_head * right_dx_tail + right_dx_head * mid_dy_tail))
            + 2 * (left_dx_head * left_dx_tail + left_dy_head * left_dy_tail)
            * (mid_dx_head * right_dy_head - mid_dy_head * right_dx_head))


def _to_squared_length(dx_head: Coordinate, dy_head: Coordinate) -> Expansion:
    dx_squared_tail, dx_squared = square(dx_head)
    dy_squared_tail, dy_squared = square(dy_head)
    return two_two_sum(dx_squared_tail, dx_squared, dy_squared_tail,
                       dy_squared)



from fractions import Fraction

from ground.hints import (Coordinate,
                          Point)


def multiply(first_start: Point,
             first_end: Point,
             second_start: Point,
             second_end: Point) -> Coordinate:
    return ((Fraction(first_end.x) - Fraction(first_start.x))
            * (Fraction(second_end.x) - Fraction(second_start.x))
            + (Fraction(first_end.y) - Fraction(first_start.y))
            * (Fraction(second_end.y) - Fraction(second_start.y)))



from fractions import Fraction
from typing import (Callable,
                    Sequence,
                    Type)

from ground.hints import Point


def centroid(point_cls: Type[Point],
             vertices: Sequence[Point],
             inverse: Callable[[int], Fraction] = Fraction(1).__truediv__
             ) -> Point:
    double_area = x_numerator = y_numerator = 0
    prev_vertex = vertices[-1]
    prev_x, prev_y = prev_vertex.x, prev_vertex.y
    for vertex in vertices:
        x, y = vertex.x, vertex.y
        area_component = prev_x * y - prev_y * x
        double_area += area_component
        x_numerator += (prev_x + x) * area_component
        y_numerator += (prev_y + y) * area_component
        prev_x, prev_y = x, y
    inverted_divisor = inverse(3 * double_area)
    return point_cls(x_numerator * inverted_divisor,
                     y_numerator * inverted_divisor)



from numbers import Real

from ground.core.shewchuk import epsilon


def to_point_point_point_determinant_error(determinant: Real) -> Real:
    return (3 + 8 * epsilon) * epsilon * abs(determinant)


def to_point_point_point_first_error(upper_bound: Real) -> Real:
    return (10 + 96 * epsilon) * epsilon * upper_bound


def to_point_point_point_second_error(upper_bound: Real) -> Real:
    return epsilon * (4 + 48 * epsilon) * upper_bound


def to_point_point_point_third_error(upper_bound: Real) -> Real:
    return epsilon * epsilon * (44 + 576 * epsilon) * upper_bound



import enum as _enum
from contextvars import ContextVar as _ContextVar
from typing import (Sequence as _Sequence,
                    Type as _Type)

from reprit.base import generate_repr as _generate_repr

from . import hints as _hints
from .core import (angular as _angular,
                   boxed as _boxed,
                   centroidal as _centroidal,
                   discrete as _discrete,
                   enums as _enums,
                   geometries as _geometries,
                   incircle as _incircle,
                   linear as _linear,
                   vector as _vector)
from .core.hints import QuaternaryPointFunction as _QuaternaryPointFunction

_QuaternaryFunction = _QuaternaryPointFunction[_hints.Coordinate]
Kind = _enums.Kind
Orientation = _enums.Orientation
Relation = _enums.Relation


@_enum.unique
class Mode(_enum.IntEnum):
    EXACT = 0
    PLAIN = 1
    ROBUST = 2


class Context:
    __slots__ = ('_box_cls', '_centroidal', '_contour_cls', '_incircle',
                 '_linear', '_mode', '_multipoint_cls', '_multipolygon_cls',
                 '_multisegment_cls', '_point_cls', '_polygon_cls',
                 '_segment_cls', '_vector')

    def __init__(self,
                 *,
                 box_cls: _Type[_hints.Box] = _geometries.Box,
                 contour_cls: _Type[_hints.Contour] = _geometries.Contour,
                 multipoint_cls: _Type[_hints.Multipoint]
                 = _geometries.Multipoint,
                 multipolygon_cls: _Type[_hints.Multipolygon]
                 = _geometries.Multipolygon,
                 multisegment_cls: _Type[_hints.Multisegment]
                 = _geometries.Multisegment,
                 point_cls: _Type[_hints.Point] = _geometries.Point,
                 polygon_cls: _Type[_hints.Polygon] = _geometries.Polygon,
                 segment_cls: _Type[_hints.Segment] = _geometries.Segment,
                 mode: Mode = Mode.EXACT) -> None:
        self._box_cls = box_cls
        self._contour_cls = contour_cls
        self._multipoint_cls = multipoint_cls
        self._multipolygon_cls = multipolygon_cls
        self._multisegment_cls = multisegment_cls
        self._point_cls = point_cls
        self._polygon_cls = polygon_cls
        self._segment_cls = segment_cls
        self._mode = mode
        self._centroidal, self._incircle, self._linear, self._vector = (
            (_centroidal.exact_context, _incircle.exact_context,
             _linear.exact_context, _vector.exact_context)
            if mode is Mode.EXACT
            else ((_centroidal.plain_context, _incircle.plain_context,
                   _linear.plain_context, _vector.plain_context)
                  if mode is Mode.PLAIN
                  else (_centroidal.robust_context, _incircle.robust_context,
                        _linear.exact_context, _vector.robust_context)))

    __repr__ = _generate_repr(__init__)

    @property
    def box_cls(self) -> _Type[_hints.Box]:
        return self._box_cls

    @property
    def contour_cls(self) -> _Type[_hints.Contour]:
        return self._contour_cls

    @property
    def cross_product(self) -> _QuaternaryFunction:
        return self._vector.cross_product

    @property
    def dot_product(self) -> _QuaternaryFunction:
        return self._vector.dot_product

    @property
    def mode(self) -> Mode:
        return self._mode

    @property
    def multipoint_cls(self) -> _Type[_hints.Multipoint]:
        return self._multipoint_cls

    @property
    def multipolygon_cls(self) -> _Type[_hints.Multipolygon]:
        return self._multipolygon_cls

    @property
    def multisegment_cls(self) -> _Type[_hints.Multisegment]:
        return self._multisegment_cls

    @property
    def point_cls(self) -> _Type[_hints.Point]:
        return self._point_cls

    @property
    def point_point_point_incircle_test(self) -> _QuaternaryFunction:
        return self._incircle.point_point_point_test

    @property
    def polygon_cls(self) -> _Type[_hints.Polygon]:
        return self._polygon_cls

    @property
    def segment_cls(self) -> _Type[_hints.Segment]:
        return self._segment_cls

    def angle_kind(self,
                   vertex: _hints.Point,
                   first_ray_point: _hints.Point,
                   second_ray_point: _hints.Point) -> Kind:
        return _angular.kind(self.dot_product, vertex, first_ray_point,
                             second_ray_point)

    def angle_orientation(self,
                          vertex: _hints.Point,
                          first_ray_point: _hints.Point,
                          second_ray_point: _hints.Point) -> Orientation:
        return _angular.orientation(self.cross_product, vertex,
                                    first_ray_point, second_ray_point)

    def contour_centroid(self,
                         vertices: _Sequence[_hints.Point]) -> _hints.Point:
        """
        Constructs centroid of a contour given its vertices.

        Time complexity:
            ``O(len(vertices))``
        Memory complexity:
            ``O(1)``

        >>> context = get_context()
        >>> Point = context.point_cls
        >>> context.contour_centroid([Point(0, 0), Point(2, 0), Point(2, 2),
        ...                           Point(0, 2)]) == Point(1, 1)
        True
        """
        return self._centroidal.contour_centroid(self.point_cls, vertices)

    def merged_box(self, first_box: _hints.Box, second_box: _hints.Box
                   ) -> _hints.Box:
        return _boxed.merge(self.box_cls, first_box, second_box)

    def multipoint_centroid(self,
                            points: _Sequence[_hints.Point]) -> _hints.Point:
        """
        Constructs centroid of a multipoint given its points.

        Time complexity:
            ``O(len(points))``
        Memory complexity:
            ``O(1)``

        >>> context = get_context()
        >>> Point = context.point_cls
        >>> context.multipoint_centroid([Point(0, 0), Point(2, 0), Point(2, 2),
        ...                              Point(0, 2)]) == Point(1, 1)
        True
        """
        return self._centroidal.multipoint_centroid(self.point_cls, points)

    def points_convex_hull(self,
                           points: _Sequence[_hints.Point]
                           ) -> _Sequence[_hints.Point]:
        """
        Constructs convex hull of points.

        Time complexity:
            ``O(points_count * log(points_count))``
        Memory complexity:
            ``O(points_count)``
        where ``points_count = len(points)``.

        >>> context = get_context()
        >>> Point = context.point_cls
        >>> (context.points_convex_hull([Point(0, 0), Point(2, 0), Point(2, 2),
        ...                              Point(0, 2)])
        ...  == [Point(0, 0), Point(2, 0), Point(2, 2), Point(0, 2)])
        True
        """
        return _discrete.to_convex_hull(self.angle_orientation, points)

    def segment_contains_point(self,
                               start: _hints.Point,
                               end: _hints.Point,
                               point: _hints.Point) -> bool:
        """
        Checks if a segment given by its endpoints contains given point.

        Time complexity:
            ``O(1)``
        Memory complexity:
            ``O(1)``

        >>> context = get_context()
        >>> Point = context.point_cls
        >>> context.segment_contains_point(Point(0, 0), Point(2, 0),
        ...                                Point(0, 0))
        True
        >>> context.segment_contains_point(Point(0, 0), Point(2, 0),
        ...                                Point(0, 2))
        False
        >>> context.segment_contains_point(Point(0, 0), Point(2, 0),
        ...                                Point(1, 0))
        True
        >>> context.segment_contains_point(Point(0, 0), Point(2, 0),
        ...                                Point(1, 1))
        False
        >>> context.segment_contains_point(Point(0, 0), Point(2, 0),
        ...                                Point(2, 0))
        True
        >>> context.segment_contains_point(Point(0, 0), Point(2, 0),
        ...                                Point(3, 0))
        False
        """
        return self._linear.containment_checker(self.cross_product, start, end,
                                                point)

    def segments_intersection(self,
                              first_start: _hints.Point,
                              first_end: _hints.Point,
                              second_start: _hints.Point,
                              second_end: _hints.Point) -> _hints.Point:
        return self._linear.intersector(
                self.cross_product, self.point_cls, first_start, first_end,
                second_start, second_end)

    def segments_relation(self,
                          test_start: _hints.Point,
                          test_end: _hints.Point,
                          goal_start: _hints.Point,
                          goal_end: _hints.Point) -> Relation:
        return self._linear.relater(self.cross_product, test_start, test_end,
                                    goal_start, goal_end)


_context = _ContextVar('context',
                       default=Context())


def get_context() -> Context:
    return _context.get()


def set_context(context: Context) -> None:
    assert isinstance(context, Context), ('Expected "{expected}" instance, '
                                          'but got "{actual}".'
                                          .format(expected=Context,
                                                  actual=context))
    _context.set(context)



from fractions import Fraction
from typing import Type

from ground.core.enums import Relation
from ground.core.hints import QuaternaryPointFunction
from ground.hints import Point


def contains_point(cross_product: QuaternaryPointFunction,
                   start: Point,
                   end: Point,
                   point: Point) -> bool:
    return (point == start or point == end
            or (_bounding_box_contains(start, end, point)
                and not cross_product(start, end, start, point)))


def intersect(cross_product: QuaternaryPointFunction,
              point_cls: Type[Point],
              first_start: Point,
              first_end: Point,
              second_start: Point,
              second_end: Point) -> Point:
    if contains_point(cross_product, first_start, first_end, second_start):
        return second_start
    elif contains_point(cross_product, first_start, first_end, second_end):
        return second_end
    elif contains_point(cross_product, second_start, second_end, first_start):
        return first_start
    elif contains_point(cross_product, second_start, second_end, first_end):
        return first_end
    else:
        scale = (cross_product(first_start, second_start, second_start,
                               second_end)
                 * (Fraction(1)
                    / cross_product(first_start, first_end, second_start,
                                    second_end)))
        return point_cls(first_start.x + (first_end.x - first_start.x) * scale,
                         first_start.y + (first_end.y - first_start.y) * scale)


def relate(cross_product: QuaternaryPointFunction,
           test_start: Point,
           test_end: Point,
           goal_start: Point,
           goal_end: Point) -> Relation:
    if test_start > test_end:
        test_start, test_end = test_end, test_start
    if goal_start > goal_end:
        goal_start, goal_end = goal_end, goal_start
    starts_equal = test_start == goal_start
    ends_equal = test_end == goal_end
    if starts_equal and ends_equal:
        return Relation.EQUAL
    test_start_cross_product = cross_product(goal_end, goal_start, goal_end,
                                             test_start)
    test_end_cross_product = cross_product(goal_end, goal_start, goal_end,
                                           test_end)
    if test_start_cross_product and test_end_cross_product:
        if (test_start_cross_product > 0) is (test_end_cross_product > 0):
            return Relation.DISJOINT
        else:
            goal_start_cross_product = cross_product(test_start, test_end,
                                                     test_start, goal_start)
            goal_end_cross_product = cross_product(test_start, test_end,
                                                   test_start, goal_end)
            if goal_start_cross_product and goal_end_cross_product:
                return (Relation.DISJOINT
                        if ((goal_start_cross_product > 0)
                            is (goal_end_cross_product > 0))
                        else Relation.CROSS)
            elif goal_start_cross_product:
                return (Relation.TOUCH
                        if test_start < goal_end < test_end
                        else Relation.DISJOINT)
            elif goal_end_cross_product:
                return (Relation.TOUCH
                        if test_start < goal_start < test_end
                        else Relation.DISJOINT)
    elif test_start_cross_product:
        return (Relation.TOUCH
                if goal_start <= test_end <= goal_end
                else Relation.DISJOINT)
    elif test_end_cross_product:
        return (Relation.TOUCH
                if goal_start <= test_start <= goal_end
                else Relation.DISJOINT)
    elif starts_equal:
        return (Relation.COMPONENT
                if test_end < goal_end
                else Relation.COMPOSITE)
    elif ends_equal:
        return (Relation.COMPOSITE
                if test_start < goal_start
                else Relation.COMPONENT)
    elif test_start == goal_end or test_end == goal_start:
        return Relation.TOUCH
    elif goal_start < test_start < goal_end:
        return (Relation.COMPONENT
                if test_end < goal_end
                else Relation.OVERLAP)
    elif test_start < goal_start < test_end:
        return (Relation.COMPOSITE
                if goal_end < test_end
                else Relation.OVERLAP)
    else:
        return Relation.DISJOINT


def _bounding_box_contains(start: Point, end: Point, point: Point) -> bool:
    start_x, start_y, end_x, end_y = start.x, start.y, end.x, end.y
    min_x, max_x = (start_x, end_x) if start_x < end_x else (end_x, start_x)
    min_y, max_y = (start_y, end_y) if start_y < end_y else (end_y, start_y)
    return min_x <= point.x <= max_x and min_y <= point.y <= max_y
